###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Relative package containing custom GUI elements."""
# TODO: Create custom QMessageBox implementation for selectable text
# TODO: Move each independent widget to its own module, grouping them in packages by functionality
# TODO: Add relative anchors for ReadmeViewer
# TODO: Zoom In/Out Buttons for LicenseViewer
# TODO: Add fading in/out animations to menus

__all__ = (
    'AppWindow',
    'ExceptionLogger',
    'ExceptionReporter',
    'ExternalTextBrowser',
    'FileContextMenu',
    'GetterApp',
    'HelpContextMenu',
    'HistoryComboBox',
    'LicenseViewer',
    'PasteLineEdit',
    'ReadmeViewer',
    'SettingsWindow',
    'Theme',
    'ToolsContextMenu',
)

from .app import *
from .menus import *
from .widgets import *
from .windows import *



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Data structures for hi_getter.network."""

__all__ = (
    'CaseInsensitiveDict',
)

from collections import OrderedDict
from collections.abc import Generator
from collections.abc import Mapping
from collections.abc import MutableMapping
from typing import Generic
from typing import TypeVar

_VT = TypeVar('_VT')


# NOTICE:
#
# Requests
# Copyright 2019 Kenneth Reitz
# Apache 2.0 License

class CaseInsensitiveDict(MutableMapping, Generic[_VT]):
    """A case-insensitive ``dict``-like object.

    Implements all methods and operations of
    ``MutableMapping`` as well as dict's ``copy``. Also
    provides ``lower_items``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive::

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()``s, the
    behavior is undefined.
    """

    def __init__(self, data=None, **kwargs) -> None:
        self._store: OrderedDict[str, _VT] = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key: str, value: _VT) -> None:
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key: str) -> _VT:
        return self._store[key.lower()][1]

    def __delitem__(self, key: str) -> None:
        del self._store[key.lower()]

    def __iter__(self) -> Generator[str]:
        return (cased_key for cased_key, mapped_value in self._store.values())

    def __len__(self) -> int:
        return len(self._store)

    def __or__(self, other: Mapping) -> 'CaseInsensitiveDict':
        if not isinstance(other, Mapping):
            return NotImplemented

        new = self.__class__(self)
        new.update(other)
        return new

    def __ror__(self, other: Mapping) -> 'CaseInsensitiveDict':
        if not isinstance(other, Mapping):
            return NotImplemented

        new = self.__class__(other)
        new.update(self)
        return new

    def __ior__(self, other: Mapping) -> 'CaseInsensitiveDict':
        self.update(other)

        return self

    def lower_items(self) -> Generator[tuple[str, _VT]]:
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lower_key, keyval[1])
            for (lower_key, keyval)
            in self._store.items()
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, Mapping):
            return NotImplemented

        other = self.__class__(other)
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self) -> 'CaseInsensitiveDict':
        """Return new :py:class:`CaseInsensitiveDict` with a copy of this instance's keys and values."""
        return self.__class__(self._store.values())

    def __repr__(self) -> str:
        return str(dict(self.items()))



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Defines the HaloInfiniteGetter HTTP Client."""

__all__ = (
    'Client',
    'TOKEN_PATH',
    'WPAUTH_PATH',
)

import json
import os
import random
import time
from pathlib import Path
from typing import Any
from typing import Final

from PySide6.QtNetwork import *
from PySide6.QtWidgets import *

from ..constants import *
from ..utils import dump_data
from ..utils import hide_windows_file
from ..utils import unique_values
from .manager import NetworkSession
from .structures import CaseInsensitiveDict
from .utils import decode_url
from .utils import guess_json_utf
from .utils import is_error_status

TOKEN_PATH:  Final[Path] = HI_CONFIG_PATH / '.token'
WPAUTH_PATH: Final[Path] = HI_CONFIG_PATH / '.wpauth'


class Client:
    """HTTP REST Client that interfaces with Halo Waypoint to get data."""

    def __init__(self, **kwargs) -> None:
        """Initializes Halo Waypoint Client

        token is first taken from token kwarg, then HI_SPARTAN_AUTH environment variable, then from user's .token file.
        wpauth is first taken from wpauth kwarg, then HI_WAYPOINT_AUTH environment variable, then from the user's .wpauth file.

        :keyword token: Token to authenticate self to 343 API.
        :keyword wpauth: Halo Waypoint authentication key, allows for creation of 343 auth tokens.
        """
        self.SVC_HOST:       str = 'svc.halowaypoint.com'
        self.WEB_HOST:       str = 'www.halowaypoint.com'
        self.parent_path:    str = '/hi/'
        self.sub_host:       str = 'gamecms-hacs-origin'  # Must be defined before session headers
        self.searched_paths: dict = {}

        self._token: str | None = kwargs.pop('token', os.getenv('HI_SPARTAN_AUTH', None))
        if self._token is None and TOKEN_PATH.is_file():
            self._token = TOKEN_PATH.read_text(encoding='utf8').strip()

        self._wpauth: str | None = kwargs.pop('wpauth', os.getenv('HI_WAYPOINT_AUTH', None))
        if self._wpauth is None and WPAUTH_PATH.is_file():
            self._wpauth = WPAUTH_PATH.read_text(encoding='utf8').strip()

        self.session: NetworkSession = NetworkSession()
        self.session.headers = CaseInsensitiveDict({
            'Accept': ', '.join(('application/json', 'text/plain', '*/*')),
            # 'Accept-Encoding': ', '.join(('gzip', 'deflate', 'br')),
            'Accept-Language': 'en-US',
            'Connection': 'keep-alive',
            'Cache-Control': 'no-cache',
            'DNT': '1',
            'Host': self.host,
            'Origin': 'https://www.halowaypoint.com',
            'Pragma': 'no-cache',
            'Referer': 'https://www.halowaypoint.com/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'Sec-GPC': '1',
            'TE': 'trailers',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:100.0) Gecko/20100101 Firefox/100.0',
        })
        self.web_session: NetworkSession = NetworkSession()
        self.web_session.headers = self.session.headers.copy() | {
            'Accept': ','.join(('text/html', 'application/xhtml+xml', 'application/xml;q=0.9', 'image/avif', 'image/webp', '*/*;q=0.8')),
            'Host': self.WEB_HOST,
            'Sec-Fetch-Dest': 'document',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-Site': 'none',
            'Sec-Fetch-User': '?1',
        }

        if self.wpauth:
            self.set_cookie('wpauth', self.wpauth)
            if not self.token:
                self.refresh_auth()

        if self.token:
            self.session.headers['x-343-authorization-spartan'] = self.token
            self.set_cookie('343-spartan-token', self.token)

    def get(self, path: str, update_auth_on_401: bool = True, **kwargs) -> QNetworkReply:
        """Get a :py:class:`Response` from HaloWaypoint.

        :param path: path to append to the API root
        :param update_auth_on_401: run self._refresh_auth if response status code is 401 Unauthorized
        :param kwargs: Key word arguments to pass to the requests GET Request.
        """
        reply: QNetworkReply = self.session.get(self.api_root + path.strip(), **kwargs)
        while not reply.isFinished():
            QApplication.processEvents()

        status_code: int = reply.attribute(QNetworkRequest.HttpStatusCodeAttribute)
        if is_error_status(status_code):
            if status_code == 401 and update_auth_on_401 and self.wpauth is not None:
                self.refresh_auth()
                reply = self.get(path, False, **kwargs)
        return reply

    def get_hi_data(self, path: str, dump_path: Path = HI_CACHE_PATH, micro_sleep: bool = True) -> dict[str, Any] | bytes | int:
        """Returns data from a path. Return type depends on the resource.

        :return: dict for JSON objects, bytes for media, int for error codes.
        """
        os_path: Path = self.os_path(path, parent=dump_path)
        data: dict[str, Any] | bytes

        if not os_path.is_file():
            reply:        QNetworkReply = self.get(path)
            encoded_data: bytes = reply.readAll().data()
            status_code:  int = reply.attribute(QNetworkRequest.HttpStatusCodeAttribute)
            content_type: str | None = reply.header(QNetworkRequest.ContentTypeHeader)

            if is_error_status(status_code):
                return status_code

            if 'json' in content_type:
                data = json.loads(encoded_data.decode(guess_json_utf(encoded_data)))
            else:
                data = encoded_data

            print(f"DOWNLOADED {path} >>> {content_type}")
            dump_data(os_path, data)

            reply.deleteLater()
            if micro_sleep:
                time.sleep(random.randint(100, 200) / 750)

        else:
            print(path)
            data = os_path.read_bytes()
            if os_path.suffix == '.json':
                data = json.loads(data.decode(guess_json_utf(data)))

        return data

    def hidden_key(self) -> str:
        """:return: The first and last 3 characters of the waypoint token, seperated by periods."""
        key = self.wpauth
        if key is not None and len(key) > 6:
            return f'{key[:3]}{"." * 50}{key[-3:]}'
        return 'None'

    def os_path(self, path, parent: Path = HI_CACHE_PATH) -> Path:
        """Translate a given GET path to the equivalent cache location."""
        return parent / self.sub_host.replace('-', '_') / self.parent_path.strip('/') / path.replace('/file/', '/').lower()

    def recursive_search(self, search_path: str) -> None:
        """Recursively get Halo Waypoint files linked to the search_path through Mapping keys."""
        if self.searched_paths.get(search_path, 0) >= 2:
            return

        data: dict[str, Any] | bytes | int = self.get_hi_data(search_path)
        if isinstance(data, (bytes, int)):
            return

        for value in unique_values(data):
            if isinstance(value, str):
                if '/' not in value:
                    continue

                end = value.split('.')[-1].lower()
                if end in ('json',):
                    path = 'progression/file/' + value
                    self.searched_paths.update({path: self.searched_paths.get(path, 0) + 1})
                    self.recursive_search(path)
                elif end in ('png', 'jpg', 'jpeg', 'webp', 'gif'):
                    self.get_hi_data('images/file/' + value)

    def refresh_auth(self) -> None:
        """Refreshes authentication to Halo Waypoint servers.

        wpauth MUST have a value for this to work. A lone 343 spartan token is not enough to generate a new one.
        """
        reply: QNetworkReply = self.web_session.get('https://www.halowaypoint.com/')

        # TODO: After moving client to separate thread, remove this and QtWidgets import.
        while not reply.isFinished():
            QApplication.processEvents()

        wpauth: str = decode_url(self.web_session.cookies.get('wpauth') or '')
        token:  str = decode_url(self.web_session.cookies.get('343-spartan-token') or '')

        if wpauth and self.wpauth != wpauth:
            self.wpauth = wpauth
        if token:
            self.token = token

    def delete_cookie(self, name: str) -> None:
        """Delete given cookie if cookie exists."""
        self.web_session.clear_cookies(self.WEB_HOST, '/', name)

    def set_cookie(self, name: str, value: str) -> None:
        """Set cookie value in Cookie jar. Defaults cookie domain as the WEB_HOST."""
        self.web_session.set_cookie(name, value, self.WEB_HOST)

    @property
    def token(self) -> str | None:
        """343 auth token to authenticate self to API endpoints."""
        return self._token

    @token.setter
    def token(self, value: str) -> None:
        value = decode_url(value)
        key_start_index = value.find('v4=')
        if key_start_index == -1:
            raise ValueError('token value is missing version identifier ("v4=") to signify start.')

        self._token = value[key_start_index:].rstrip()
        self.set_cookie('343-spartan-token', self._token)
        self.session.headers['x-343-authorization-spartan'] = self._token
        hide_windows_file(TOKEN_PATH, unhide=True)
        TOKEN_PATH.write_text(self._token, encoding='utf8')
        hide_windows_file(TOKEN_PATH)

    @token.deleter
    def token(self) -> None:
        self._token = None
        self.delete_cookie('343-spartan-token')
        if 'x-343-authorization-spartan' in self.session.headers:
            self.session.headers.pop('x-343-authorization-spartan')
        if TOKEN_PATH.is_file():
            TOKEN_PATH.unlink()

    @property
    def wpauth(self) -> str | None:
        """Halo Waypoint auth key to create 343 spartan tokens."""
        return self._wpauth

    @wpauth.setter
    def wpauth(self, value: str) -> None:
        value = decode_url(value)
        self._wpauth = value.split(':')[-1].strip()
        self.set_cookie('wpauth', self._wpauth)
        hide_windows_file(WPAUTH_PATH, unhide=True)
        WPAUTH_PATH.write_text(self._wpauth, encoding='utf8')
        hide_windows_file(WPAUTH_PATH)

    @wpauth.deleter
    def wpauth(self) -> None:
        self._wpauth = None
        self.delete_cookie('wpauth')
        if WPAUTH_PATH.is_file():
            WPAUTH_PATH.unlink()

    @property
    def api_root(self) -> str:
        """Root of sent API requests."""
        return f'https://{self.host}{self.parent_path}'

    @property
    def host(self) -> str:
        """Host to send to."""
        return f'{self.sub_host}.{self.SVC_HOST}'

    @host.setter
    def host(self, value: str) -> None:
        self.sub_host = value



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Version information."""
# Store the version here so:
# 1) We don't load dependencies by storing it in __init__.py
# 2) We can load it in setup.cfg
# 3) We can import it into modules

__version_info__ = (0, 10, 0, 'alpha', 2)
"""Major, Minor, Micro, Release level, Serial in respective order."""


def _stringify(major: int, minor: int, micro: int = 0, releaselevel: str = 'final', serial: int = 0, **kwargs) -> str:
    """Stringifies a version number based on version info given. Follows :pep:`440`.

    Releaselevel is the status of the given version, NOT the project itself.
    All versions of an alpha or beta should be a 'final' releaselevel.

    Serial is only taken into account if releaselevel is not 'final' or 'release'.

    For developmental releases, post releases, and local release specifications, see
    https://www.python.org/dev/peps/pep-0440/#normalization

    | Ex: (2021, 9) -> 2021.9
    | Ex: (0, 3, 2, 'beta') -> 0.3.2b
    | Ex: (1, 0, 0, 'release') -> 1.0
    | Ex: (3, 10, 0, 'candidate', 0) -> 3.10rc
    | Ex: (3, 9, 1, 'alpha', 3) -> 3.9.2a3
    | Ex: (3, 9, 1, dev=2) -> 3.9.2.dev2
    | Ex: (3, 9, 2, 'preview', 3, post=0, post_implicit=True, dev=5) -> 3.9.2pre3-0.dev5
    | Ex: (1, 0, local='ubuntu', local_ver='2', local_ver_sep='-') -> 1.0+ubuntu-2

    :param major: First and most important version number.
    :param minor: Second and less important version number.
    :param micro: Last and least important version number.
    :param releaselevel: Status of current version.
    :param serial: Separate version number of an Alpha/Beta for an upcoming release version.
    :keyword local: Local release type, ex: 'ubuntu'
    :keyword local_ver: Local release version
    :keyword local_ver_sep: Local release seperator
    :keyword pre_sep: Pre-release seperator
    :keyword pre_ver_sep: Pre-release version seperator
    :keyword post: Post-release version
    :keyword post_implicit: Post-release version
    :keyword post_sep: Post-release seperator
    :keyword post_ver_sep: Post-release version seperator
    :keyword dev: Developmental release version
    :keyword dev_sep: Developmental release seperator
    :keyword dev_post: Developmental Post-release version
    :keyword dev_post_sep: Developmental Post-release seperator
    :keyword dev_post_ver_sep: Developmental Post-release version seperator
    :return: String representation of version number.
    :raises TypeError: Version numbers are not integers.
    :raises ValueError: Version separators are not in the allowed chars.
    """
    (local, local_ver, local_ver_sep, pre_sep, pre_ver_sep,
     post, post_spelling, post_implicit, post_sep, post_ver_sep,
     dev, dev_sep, dev_post, dev_post_sep, dev_post_ver_sep) = (
        kwargs.pop('local', None),
        kwargs.pop('local_ver', 0),
        kwargs.pop('local_ver_sep', '.'),
        kwargs.pop('pre_sep', ''),
        kwargs.pop('pre_ver_sep', ''),
        kwargs.pop('post', None),
        kwargs.pop('post_spelling', 'post'),
        kwargs.pop('post_implicit', False),
        kwargs.pop('post_sep', ''),
        kwargs.pop('post_ver_sep', ''),
        kwargs.pop('dev', None),
        kwargs.pop('dev_sep', ''),
        kwargs.pop('dev_post', None),
        kwargs.pop('dev_post_sep', ''),
        kwargs.pop('dev_post_ver_sep', ''),
    )

    releaselevel = releaselevel.strip()
    separators:     tuple[str, ...] = ('', '.', '-', '_')
    post_spellings: tuple[str, ...] = ('post', 'rev', 'r')
    release_levels: tuple[str, ...] = ('a', 'b', 'c', 'rc', 'pre', 'alpha', 'beta', 'candidate', 'preview', 'final', 'release')

    for attr in ('major', 'minor', 'micro', 'local_ver', 'post', 'dev', 'dev_post'):
        if vars()[attr] is not None and not isinstance(vars()[attr], int):
            raise TypeError(f'Argument "{attr}" should be of type {int}')  # pragma: no cover
    if (False in (sep in separators for sep in (pre_sep, pre_ver_sep, post_sep, post_ver_sep, dev_sep, dev_post_sep, dev_post_ver_sep)) or
            local_ver_sep not in separators[1:]):
        raise ValueError(f'A separator given is not in allowed separators {separators}')
    if post_spelling not in post_spellings:
        raise ValueError(f'Post-release spelling not allowed as "{post_spelling}"')
    if releaselevel not in release_levels:
        raise ValueError(f'Release level "{releaselevel}" is not in known release levels')

    v_number: str = f'{major}.{minor}'
    v_number += f'.{micro}' if micro else ''

    if releaselevel and releaselevel not in ('final', 'release'):
        if releaselevel in ('candidate', 'c'):
            releaselevel = 'rc'
        if releaselevel in ('alpha', 'beta'):
            releaselevel = releaselevel[0]
        if releaselevel == 'preview':
            releaselevel = 'pre'
        v_number += pre_sep
        v_number += releaselevel if len(releaselevel) <= 3 else releaselevel[0]
        v_number += f'{pre_ver_sep}{serial}' if serial else ''

    if post is not None:
        sep = f'{post_sep}{post_spelling}' if not post_implicit else '-'
        v_number += f'{sep}{post_ver_sep if post else ""}{post if post or sep == "-" else ""}'

    if dev is not None:
        v_number += f'{dev_sep}dev{dev if dev else ""}'

    if dev_post is not None:
        sep = f'{dev_post_sep + post_spelling}' if not post_implicit else '-'
        v_number += f'{sep}{dev_post_ver_sep if dev_post else ""}{dev_post if dev_post or sep == "-" else ""}'

    if local is not None:
        v_number += f'+{local + local_ver_sep}{local_ver}'

    return v_number


__version__ = _stringify(*__version_info__)
"""String representation of version number."""



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing miscellaneous :py:class:`QWidget` Widgets."""

__all__ = (
    'ExceptionReporter',
    'ExceptionLogger',
    'ExternalTextBrowser',
    'HistoryComboBox',
    'LicenseViewer',
    'PasteLineEdit',
    'ReadmeViewer',
)

import traceback
import webbrowser
from collections import defaultdict
from collections import namedtuple
from collections.abc import Callable
from collections.abc import Sequence
from datetime import datetime
from importlib.metadata import metadata
from types import TracebackType
from typing import Any
from typing import Final
from typing import Optional

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from .._version import __version__
from ..constants import *
from ..events import EventBus
from ..exceptions import ExceptionEvent
from ..models import DeferredCallable
from ..models import DistributedCallable
from ..network import encode_url_params
from ..utils import current_requirement_licenses
from .app import app
from .utils import delete_layout_widgets
from .utils import init_objects
from .utils import PARENT_PACKAGE
from .utils import scroll_to_top


class ExceptionReporter(QWidget):
    """A :py:class:`QWidget` that displays logged exceptions and their traceback."""

    def __init__(self, logger: 'ExceptionLogger') -> None:
        super().__init__()
        EventBus['exceptions'].subscribe(DeferredCallable(self.reload_exceptions), ExceptionEvent)

        self.selected: int = 0
        self.logger: ExceptionLogger = logger
        self.setWindowTitle(app().translator('gui.exception_reporter.title'))
        self.setWindowIcon(self.logger.icon())
        self.resize(QSize(750, 400))
        self._init_ui()

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)
        self.left_panel = QFrame(self)
        self.right_panel = QFrame(self)

        left_label = QLabel(self.left_panel)
        right_label = QLabel(self.right_panel)
        clear_all_button = QPushButton(self.left_panel)
        self.trace_back_viewer = ExternalTextBrowser(self)
        self.clear_button = QPushButton(self.right_panel)
        self.report_button = QPushButton(self.right_panel)

        # Define the scroll area/widget for the left panel
        self.scroll_area = QScrollArea(self)
        self.scroll_widget = QWidget(self.scroll_area)
        self.scroll_widget.setLayout(QVBoxLayout())
        self.scroll_widget.layout().setAlignment(Qt.AlignTop)
        self.scroll_area.setWidget(self.scroll_widget)

        init_objects({
            self.scroll_area: {
                'widgetResizable': True
            },
            clear_all_button: {
                'clicked': self.clear_all_exceptions
            },
            self.trace_back_viewer: {
                'disabled': True,
                'font': QFont('consolas', 10),
                'lineWrapMode': QTextEdit.NoWrap
            },
            self.clear_button: {
                'disabled': True,
                'size': {'maximum': (self.report_button.width() // 1, None)},
                'clicked': self.clear_current_exception
            },
            self.report_button: {
                'disabled': True,
                'clicked': self.report_current_exception
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.exception_reporter.title',

            # Labels
            left_label.setText: 'gui.exception_reporter.exception_list',
            right_label.setText: 'gui.exception_reporter.traceback_label',

            # Buttons
            clear_all_button.setText: 'gui.exception_reporter.clear_all',
            self.clear_button.setText: 'gui.exception_reporter.clear',
            self.report_button.setText: 'gui.exception_reporter.report_clear'
        })

        layout.addWidget(self.left_panel)
        layout.addWidget(self.right_panel)
        self.left_panel.setLayout(QVBoxLayout())
        self.right_panel.setLayout(QVBoxLayout())
        buttons = QHBoxLayout()

        self.left_panel.layout().addWidget(left_label)
        self.left_panel.layout().addWidget(self.scroll_area)
        self.left_panel.layout().addWidget(clear_all_button)

        self.right_panel.layout().addWidget(right_label)
        self.right_panel.layout().addWidget(self.trace_back_viewer)
        self.right_panel.layout().addLayout(buttons)
        buttons.addWidget(self.report_button)
        buttons.addWidget(self.clear_button)

    def report_current_exception(self) -> None:
        """Report the current exception to the exception logger."""
        exc_name = type(self.logger.exception_log[self.selected].exception).__name__
        exc_msg = str(self.logger.exception_log[self.selected].exception).rstrip(".")
        exc_tb = traceback.format_tb(self.logger.exception_log[self.selected].traceback)[0].strip()

        base: str = 'https://github.com/Cubicpath/HaloInfiniteGetter/issues/new'
        params = {
            'template': 'bug-report.yaml',
            'assignees': 'Cubicpath',
            'labels': 'bug',
            'projects': 'Cubicpath/HaloInfiniteGetter/2',
            'title': f'[Bug]: ({exc_name}) {exc_msg}',
            'version': f'v{__version__}',
            'logs': f'{exc_name}: {exc_msg}\n\n{exc_tb}'
        }
        webbrowser.open(f'{base}?{encode_url_params(params)}')

    def disable_exception_widgets(self) -> None:
        """Disables the widgets on the right pane."""
        self.trace_back_viewer.clear()
        self.trace_back_viewer.setDisabled(True)
        self.report_button.setDisabled(True)
        self.clear_button.setDisabled(True)

    def clear_all_exceptions(self) -> None:
        """Clear all exceptions from the list."""
        self.logger.clear_exceptions()
        self.disable_exception_widgets()
        delete_layout_widgets(self.scroll_widget.layout())

    def clear_current_exception(self) -> None:
        """Clears the currently selected exception and removes it from the log."""
        if (item := self.scroll_widget.layout().takeAt(self.selected)) is not None:
            self.logger.remove_exception(self.selected)
            item.widget().deleteLater()
        self.reload_exceptions()
        self.disable_exception_widgets()

    def reload_exceptions(self) -> None:
        """Load the exceptions from the logger."""
        delete_layout_widgets(self.scroll_widget.layout())
        for i, error in enumerate(self.logger.exception_log):
            button = QPushButton(f'{type(error.exception).__name__}: {error.exception}', self.scroll_widget)
            button.clicked.connect(DeferredCallable(setattr, self, 'selected', i))
            button.clicked.connect(DistributedCallable((
                self.clear_button.setDisabled,
                self.report_button.setDisabled,
                self.trace_back_viewer.setDisabled
            ), False))
            button.clicked.connect(DeferredCallable(self.trace_back_viewer.setText, DeferredCallable(
                app().translator, 'gui.exception_reporter.traceback_view',
                type(error.exception).__name__, error.exception, traceback.format_tb(error.traceback)[0]
            )))

            button.setStyleSheet("text-align:left;")
            button.setMaximumWidth((self.size().width() // 3) - 50)

            self.scroll_widget.layout().addWidget(button)

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Resizes the left panel to better fit the window."""
        self.left_panel.setMaximumWidth(event.size().width() // 3)
        for i in range(self.scroll_widget.layout().count()):
            self.scroll_widget.layout().itemAt(i).widget().setMaximumWidth((event.size().width() // 3) - 50)
        event.accept()


class ExceptionLogger(QPushButton):
    """A :py:class:`QPushButton` that logs exceptions to the event bus."""
    LoggedException: tuple[int, Exception, TracebackType, datetime] = namedtuple('LoggedException', [
        'severity', 'exception', 'traceback', 'timestamp'
    ], defaults=[None])
    """A named tuple that contains the severity of the exception, the exception itself, and an optional traceback."""

    level_icon_list: list = [
        QStyle.SP_MessageBoxInformation,  # 0, Not a concern
        QStyle.SP_MessageBoxWarning,      # 1, Warning
        QStyle.SP_MessageBoxCritical      # 2, Error
    ]

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the ExceptionLogger."""
        super().__init__(*args, **kwargs)
        EventBus['exceptions'].subscribe(self.on_exception, ExceptionEvent)

        self.exception_log:    list[ExceptionLogger.LoggedException] = []
        self.reporter:         ExceptionReporter = ExceptionReporter(self)
        self.severity:         int = 0

        self.reporter.setMinimumWidth(300)

    def clear_exceptions(self) -> None:
        """Clear the exception log and disable the button."""
        self.exception_log.clear()
        self.severity = 0
        self.setText('')

    def remove_exception(self, index: int) -> None:
        """Remove an exception from the log and update the current severity."""
        if self.exception_log:
            self.exception_log.pop(index)

        if len(self.exception_log) == 0:
            self.clear_exceptions()
        else:
            logged = len(self.exception_log)
            self.setText(f'({logged})' if logged < 10 else '(9+)')
            self.severity = self.exception_log[0].severity

    def on_exception(self, event: ExceptionEvent) -> None:
        """Update the exception log and change set the max level."""
        if isinstance(event.exception, Warning) and self.severity < 1:
            level = 1
        else:
            level = 2
        if self.severity < level:
            self.severity = level

        self.exception_log.append(self.LoggedException(level, event.exception, event.traceback, datetime.now()))
        self.sort_exceptions()

        logged = len(self.exception_log)
        self.setText(f'({logged})' if logged < 10 else '(9+)')

    def sort_exceptions(self) -> None:
        """Sort the exception log by severity."""
        if self.exception_log:
            self.exception_log = list(reversed(sorted(self.exception_log, key=lambda x: x.severity)))
            self.severity = self.exception_log[0].severity
        else:
            self.severity = 0

    @property
    def severity(self) -> int:
        """Get the max level of the exception log."""
        return self._severity

    @severity.setter
    def severity(self, value: int) -> None:
        """Set the max level of the exception log and update the icon."""
        self._severity = value
        self.setIcon(self.style().standardIcon(self.level_icon_list[self.severity]))
        self.reporter.setWindowIcon(self.icon())


class ExternalTextBrowser(QTextBrowser):
    """:py:class:`QTextBrowser` with ability to map keys to :py:class:`Callable`'s.

    Also supports external image loading and caching.
    """
    remote_image_cache: dict[str, bytes] = {}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # default factory producing empty DeferredCallables
        self.key_callable_map: defaultdict[int, Callable] = defaultdict(DeferredCallable)
        self.cached_text: str = ''
        self.cached_type: str = ''

    def hot_reload(self) -> None:
        """Reload cached text using the cached type's function."""
        # Remember scroll position
        scroll = self.verticalScrollBar().sliderPosition()

        # Reload all resources and replace text with cached text
        self.clear()
        match self.cached_type:
            case 'markdown':
                self.setMarkdown(self.cached_text)
            case 'html':
                self.setHtml(self.cached_text)
            case 'text':
                self.setPlainText(self.cached_text)
        self.verticalScrollBar().setSliderPosition(scroll)

    def set_hot_reloadable_text(self, text: str, text_type: str) -> None:
        """Set text that is designated to be hot-reloadable."""
        self.cached_text = text
        self.cached_type = text_type
        self.hot_reload()

    def loadResource(self, resource_type: QTextDocument.ResourceType, url: QUrl) -> Any:
        """Load a resource from an url.

        If resource type is an image and the url is external, download it using requests and cache it.
        """
        if resource_type == QTextDocument.ResourceType.ImageResource and not url.isLocalFile():
            image:      QImage = QImage()
            url_string: str = url.toDisplayString()
            if url_string not in self.remote_image_cache:
                reply = app().session.get(url)

                def handle_reply():
                    data: bytes = reply.readAll()
                    image.loadFromData(data)
                    self.remote_image_cache[url_string] = data

                    if self.cached_type:
                        self.hot_reload()
                    reply.deleteLater()

                reply.finished.connect(handle_reply)
            else:
                image.loadFromData(self.remote_image_cache[url_string])
            return image
        else:
            return super().loadResource(int(resource_type), url)

    def setLineWrapMode(self, mode: int | QTextEdit.LineWrapMode) -> None:
        """Set the line wrap mode. Allows use of ints."""
        super().setLineWrapMode(QTextEdit.LineWrapMode(mode))

    # noinspection PyTypeChecker
    def connect_key_to(self, key: Qt.Key, func: Callable) -> None:
        """Connect a :py:class:`Callable` to a key press."""
        self.key_callable_map[int(key)] = func

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Execute :py:class:`Callable` mapped to the key press."""
        self.key_callable_map[event.key()]()
        event.accept()


class PasteLineEdit(QLineEdit):
    """A :py:class:`QLineEdit` with an added paste listener."""
    pasted = Signal(name='pasted')

    def __init__(self, *args, pasted: Optional[Callable] = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if pasted is not None:
            self.pasted.connect(pasted)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Call self.pasted on paste."""
        if event.matches(QKeySequence.Paste):
            self.pasted.emit()
        event.accept()


class HistoryComboBox(QComboBox):
    """Editable :py:class:`QComboBox` acting as a history wrapper over :py:class:`BetterLineEdit`; has no duplicate values."""
    line_edit_class = PasteLineEdit

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setEditable(True)
        self.setDuplicatesEnabled(False)
        self.setLineEdit(self.line_edit_class(parent=self))

    # noinspection PyTypeChecker
    def addItem(self, text: str, **kwargs) -> None:
        """Filters already-present strings from being added using addItem.

        addItem(self,
        icon: Union[PySide6.QtGui.QIcon, PySide6.QtGui.QPixmap],
        text: str,
        userData: Any = Invalid(typing.Any)
        ) -> None
        addItem(self,
        text: str,
        userData: Any = Invalid(typing.Any)
        ) -> None"""
        result = self.findText(text, Qt.MatchFlag.MatchFixedString)
        if result != -1:
            self.removeItem(result)

        super().addItem(text, **kwargs)

    def addItems(self, texts: Sequence[str], **_) -> None:
        """self.addItem for text in texts."""
        for text in texts:
            self.addItem(text)


class LicenseViewer(QWidget):
    """Widget that formats and shows the project's (and all of its requirements') license files."""
    LICENSE_DATA: Final[dict[str, tuple[str, str]]] = current_requirement_licenses(PARENT_PACKAGE)

    def __init__(self, *args, **kwargs) -> None:
        """Create a new LicenseViewer. Gets license from the HI_RESOURCE_PATH/LICENSE file

        Has a fixed size of 750x380.
        """
        super().__init__(*args, **kwargs)
        self.setWindowTitle(app().translator('gui.license_viewer.title'))
        self.setWindowIcon(app().icon_store['copyright'])
        self.resize(QSize(750, 550))
        self.current_license_index = 0

        self.license_label:       QLabel
        self.license_index_label: QLabel
        self.license_text_edit:   ExternalTextBrowser
        self.next_license_button: QPushButton
        self.prev_license_button: QPushButton
        self._init_ui()

    def _init_ui(self) -> None:
        self.license_label:       QLabel = QLabel(self)
        self.license_index_label: QLabel = QLabel(f'{self.current_license_index + 1} of {len(self.LICENSE_DATA)}', self)
        self.license_text_edit:   ExternalTextBrowser = ExternalTextBrowser(self)
        self.next_license_button: QPushButton = QPushButton(app().translator('gui.license_viewer.next'), clicked=self.next_license)
        self.prev_license_button: QPushButton = QPushButton(app().translator('gui.license_viewer.previous'), clicked=self.prev_license)

        self.license_text_edit.connect_key_to(Qt.Key_Left, self.prev_license)
        self.license_text_edit.connect_key_to(Qt.Key_Right, self.next_license)
        self.view_package(PARENT_PACKAGE)

        layout = QVBoxLayout(self)
        top = QHBoxLayout()

        layout.addLayout(top)
        top.addWidget(self.license_label)
        top.addWidget(self.prev_license_button)
        top.addWidget(self.next_license_button)
        top.addWidget(self.license_index_label)
        layout.addWidget(self.license_text_edit)

        cursor = self.license_text_edit.textCursor()
        cursor.clearSelection()
        cursor.select(QTextCursor.SelectionType.Document)
        self.next_license_button.setMaximumWidth(100)
        self.prev_license_button.setMaximumWidth(100)
        self.license_index_label.setMaximumWidth(50)
        self.license_text_edit.setFont(QFont('consolas', 11))
        self.license_text_edit.setOpenExternalLinks(True)

    def next_license(self) -> None:
        """View the next license."""
        self.current_license_index += 1
        if self.current_license_index + 1 > len(self.LICENSE_DATA):
            self.current_license_index = 0
        self.view_current_index()

    def prev_license(self) -> None:
        """View the previous license."""
        self.current_license_index -= 1
        if self.current_license_index < 0:
            self.current_license_index = len(self.LICENSE_DATA) - 1
        self.view_current_index()

    def view_current_index(self) -> None:
        """Views the license data at the current index."""
        self.view_package(tuple(self.LICENSE_DATA)[self.current_license_index])

    def view_package(self, package: str) -> None:
        """Views the license data of the given package name."""
        license_text = self.LICENSE_DATA[package][1] or app().translator('gui.license_viewer.not_found')
        self.current_license_index = tuple(self.LICENSE_DATA).index(package)
        self.license_label.setText(f'{package} -- "{self.LICENSE_DATA[package][0]}" {app().translator("gui.license_viewer.license")}')
        self.license_index_label.setText(f'{self.current_license_index + 1} of {len(self.LICENSE_DATA)}')

        output = license_text
        replaced = set()
        for match in HI_URL_PATTERN.finditer(license_text):
            match = match[0]
            if match not in replaced:
                output = output.replace(match, f'<a href="{match}" style="color: #2A5DB0">{match}</a>')
                replaced.add(match)

        stripped_output = ''
        for line in output.splitlines():
            stripped_output += line.strip() + '\n'
        stripped_output = stripped_output.strip()
        self.license_text_edit.setHtml(
            f'<body style="white-space: pre-wrap">'
            f'<center>{stripped_output}</center>'
            f'</body>'
        )

        scroll_to_top(self.license_text_edit)


class ReadmeViewer(QWidget):
    """Widget that formats and shows the project's README.md, stored in the projects 'Description' metadata tag."""
    README_TEXT: Final[str] = metadata(PARENT_PACKAGE)['Description']

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWindowTitle(app().translator('gui.readme_viewer.title'))
        self.setWindowIcon(self.style().standardIcon(QStyle.SP_DialogApplyButton))
        self.resize(QSize(750, 750))
        self.readme_viewer: ExternalTextBrowser
        self._init_ui()

    def _dummy_func(self) -> None:
        """Must exist otherwise ReadmeViewer instances will be garbage collected through Context Menu deletion. Don't ask, just accept."""

    def _init_ui(self) -> None:
        self.readme_viewer = ExternalTextBrowser(self)
        close_button = QPushButton("Close", self, clicked=self.close)

        self.readme_viewer.connect_key_to(Qt.Key_Any, self._dummy_func)  # Refer to self._dummy_func

        layout = QVBoxLayout(self)
        self.setLayout(layout)

        layout.addWidget(self.readme_viewer)
        layout.addWidget(close_button)

        self.readme_viewer.setOpenExternalLinks(True)
        self.readme_viewer.set_hot_reloadable_text(self.README_TEXT, 'markdown')
        self.readme_viewer.setFont(QFont(self.readme_viewer.font().family(), 10))
        close_button.setMinimumHeight(40)
        close_button.setFont(QFont(close_button.font().family(), 16))

    def closeEvent(self, event: QCloseEvent) -> None:
        """Manually signal the readme_viewer for garbage collection."""
        self.readme_viewer.deleteLater()
        event.accept()



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Neutral namespace for constants. Meant to be star imported."""

__all__ = (
    'BYTE_UNITS',
    'HI_CACHE_PATH',
    'HI_CONFIG_PATH',
    'HI_PATH_PATTERN',
    'HI_RESOURCE_PATH',
    'HI_SAMPLE_RESOURCE',
    'HI_URL_PATTERN',
)

import re
from pathlib import Path
from typing import Final

# Mappings

BYTE_UNITS = {'Bytes': 1024**0, 'KiB': 1024**1, 'MiB': 1024**2, 'GiB': 1024**3, 'TiB': 1024**4}

# Strings

HI_SAMPLE_RESOURCE: Final[str] = 'Progression/file/Calendars/Seasons/SeasonCalendar.json'
"""Example resource. Is pre-filled in the search bar."""

# Paths

HI_CACHE_PATH:      Final[Path] = Path.home() / '.cache/hi_getter'
"""Directory containing cached API results."""

HI_CONFIG_PATH:     Final[Path] = Path.home() / '.config/hi_getter'
"""Directory containing user configuration data."""

HI_RESOURCE_PATH:   Final[Path] = Path(__file__).parent / 'resources'
"""Directory containing application resources."""

# Patterns

HI_PATH_PATTERN:    Final[re.Pattern] = re.compile(r'([A-Z]:\\)*([\w\-_.]+[/\\])+[\w\-_]*\.\w+|[/\\]\.\w+')
"""Regex pattern for finding a resource path. Finds quoted substrings with at least one folder name and file name (with a file extension)."""

HI_URL_PATTERN:     Final[re.Pattern] = re.compile(
    r'(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|'
    r'[a-z0-9.\-]+[.][a-z]{2,}/)(?:[^\s()<>{}\[\]]+|'
    r'\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|'
    r'\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|'
    r'\([^\s]+?\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’])|'
    r'(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.][a-z]{2,}\b/?(?!@))')
"""Regex pattern for finding URLs. Derived from https://gist.github.com/gruber/8891611."""



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module used for Event subscription."""

__all__ = (
    'Event',
    'EventBus',
)

from collections import defaultdict
from collections.abc import Callable
from typing import Generic
from typing import Optional
from typing import overload
from typing import TypeAlias
from typing import TypeVar

from . import utils


class Event:
    """Normal event with no special abilities.

    You can also use the >> operator to fire an :py:class:`Event` on an :py:class:`EventBus`. Ex::

        Event() >> EventBus['foo']
    """
    __slots__ = ()

    def __repr__(self) -> str:
        values = {attr: val for attr, val in ((attr, getattr(self, attr)) for attr in self.__slots__)}
        return f'<"{self.name}" Event {values=}>' if self.__slots__ else f'<Empty {self.name}>'

    def __rshift__(self, __bus: 'EventBus', /) -> None:
        """Syntax sugar for __bus.fire(event)"""
        __bus.fire(event=self)

    def __rlshift__(self, __bus: 'EventBus', /) -> None:
        """Right shift if the EventBus.__lshift__ dunder is not working."""
        self >> __bus

    @property
    def name(self) -> str:
        """Name of event.

        Defaults to class name.
        """
        return type(self).__name__

    @property
    def description(self) -> str:
        """Short description of the event.

        Defaults to the first line of the nearest type doc in the mro.
        """
        doc = self.__doc__
        if doc is None and isinstance(self, Event) and type(self) is not Event:
            doc = utils.get_parent_doc(type(self))
        return doc.splitlines()[0]


_ET_co = TypeVar('_ET_co', bound=Event, covariant=True)  # Bound to Event. Can use Event subclass instances in place of Event instances.
_EventPredicate: TypeAlias = Callable[[_ET_co], bool]
_EventRunnable:  TypeAlias = Callable[[_ET_co], None]


class _Subscribers(defaultdict[type[Event], list[tuple[
    _EventRunnable,          # Callable to run
    _EventPredicate | None   # Optional predicate to run callable
]]], Generic[_ET_co]):
    """Class which holds the event subscribers for an :py:class:`EventBus`."""

    def __init__(self) -> None:
        super().__init__(list)

    def __repr__(self) -> str:
        """Amount of subscribers for every event, encased in parentheses."""
        repr_: str = ''
        for event in self:
            repr_ += f'{event.__name__}[{len(self[event])}], '
        return f'({repr_.rstrip()})'

    def add(self, event: type[_ET_co], callable_pair: tuple[_EventRunnable, _EventPredicate | None]) -> None:
        """Add a callable pair to an Event subscriber list."""
        if not issubclass(event, Event):
            raise TypeError(f'event is not subclass to {Event}.')
        if not callable(callable_pair[0]):
            raise TypeError('subscriber is not callable.')
        if callable_pair[1] is not None and not callable(callable_pair[1]):
            raise TypeError('subscriber predicate is defined but not callable.')

        self[event].append(callable_pair)


class _EventBusMeta(type):
    """Metaclass for :py:class:`EventBus` type.

    Maps :py:class:`EventBus` objects to :py:class:`str` ids
    and allows accessing those ids using subscripts on :py:class:`EventBus`.
    """
    _id_bus_map: dict[str, 'EventBus'] = {}

    @classmethod
    @overload
    def __getitem__(mcs, _: type[Event] | str, /) -> Optional['EventBus']:
        """Exist to support Generic Typing on :py:class:`EventBus`."""

    @classmethod
    def __getitem__(mcs, id: str) -> Optional['EventBus']:
        return mcs._id_bus_map.get(id.lower())

    @classmethod
    def __setitem__(mcs, id: str, bus: 'EventBus') -> None:
        """Set an :py:class:`EventBus` from the bus map."""
        if not isinstance(id, str):
            raise TypeError(f'parameter {id=} is not of type {str}.')
        if not isinstance(bus, EventBus):
            raise TypeError(f'parameter {bus=} is not of type {EventBus}.')
        mcs._id_bus_map[id.lower()] = bus

    @classmethod
    def __delitem__(mcs, id: str) -> None:
        """Delete an :py:class:`EventBus` from the bus map."""
        del mcs._id_bus_map[id.lower()]


class EventBus(Generic[_ET_co], metaclass=_EventBusMeta):
    """Object that keeps track of all :py:class:`Callable` subscriptions to :py:class:`Event`'s.

    An Event Bus is an event-driven structure that stores both events and functions. When an
    event is fired, all functions "subscribed" to the event or any of its parent events are called with
    the event as a parameter.

    All :py:class:`EventBus`'s are stored in the class with a unique id.
    You can access created :py:class:`EventBus`'s with subscripts. ex::

        [Python Interactive Prompt]
            EventBus['foo'] -> None
            EventBus('foo') -> EventBus object at 0x000000001
            EventBus['foo'] -> EventBus object at 0x000000001
            del EventBus['foo']
            EventBus['foo'] -> None

    You can also use the << operator to fire an event, or an Event's >> operator. ex::

        [Python Interactive Prompt]
            # These fire an event
            EventBus['foo'] << Event()
            Event() >> EventBus['foo']

            # This raises an error:
            Event() << EventBus['foo']
    """

    def __init__(self, __id: str | None = None, /) -> None:
        """Create a new :py:class:`EventBus` object with a unique id.

        All ids are transformed to lowercase in the :py:class:`_EventBusMeta` id map.

        :param __id: id to register this instance as.
        :raises KeyError: When a bus with the given id already exists.
        """
        if __id is None or type(self)[__id] is None:
            self.id = __id
            self._subscribers = _Subscribers()
            if __id is not None:
                type(self)[__id] = self
        else:
            raise KeyError(f'EventBus id "{__id}" is already registered in {type(type(self)).__name__}')

    def __lshift__(self, __event: _ET_co | type[_ET_co], /) -> None:
        """Syntax sugar for self.fire(event)"""
        self.fire(__event)

    def __repr__(self) -> str:
        return f'<{type(self).__name__} id={self.id!r}; Subscribers={self._subscribers!r}>'

    def clear(self, event: type[_ET_co] | None = None) -> None:
        """Clear event _subscribers of a given type.

        None is treated as a wildcard and deletes ALL event _subscribers.

        :param event: Event type to clear.
        """
        if event is None:
            self._subscribers.clear()
        else:
            self._subscribers.pop(event)

    def fire(self, event: _ET_co | type[_ET_co]) -> None:
        """Fire all :py:class:`Callables` subscribed to the :py:class:`Event`'s :py:class:`type`.

        :py:class:`Event` subclasses call their parent's callables as well.
        Ex: ChildEvent will fire ParentEvent, but ParentEvent will not fire ChildEvent.

        If event is a :py:class:`type`, it is instantiated with no arguments.

        :param event: Event object passed to callables as an argument.
        """
        # Transform Event types to their default instances
        if isinstance(event, type) and issubclass(event, Event):
            event = event()

        # Run all current and parent event callables
        for event_type in self._subscribers:
            if isinstance(event, event_type):
                for e_callable_pair in self._subscribers[event_type]:
                    # Check predicate if one is given
                    if e_callable_pair[1] is None or e_callable_pair[1](event):

                        # Finally, call
                        e_callable_pair[0](event)

    def subscribe(self, __callable: _EventRunnable, /, event: type[_ET_co], event_predicate: _EventPredicate | None = None) -> None:
        """Subscribe a :py:class:`Callable` to an :py:class:`Event` type.

        By default, every time an :py:class:`Event` is fired, it will call the callable with the event as an argument.

        :param __callable: Callable to run.
        :param event: Event to subscribe to.
        :param event_predicate: Predicate to validate before running callable.
        :raises TypeError: If the given arguments are not valid.
        """
        callable_pair = (__callable, event_predicate)
        self._subscribers.add(event, callable_pair)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing utils for GUI elements."""

__all__ = (
    'delete_layout_widgets',
    'icon_from_bytes',
    'init_objects',
    'PARENT_PACKAGE',
    'scroll_to_top',
    'set_or_swap_icon',
)

from collections.abc import Iterable
from collections.abc import Sequence
from typing import Any
from typing import Final

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..lang import Translator

PARENT_PACKAGE: Final[str] = __package__.split('.', maxsplit=1)[0]


def _return_arg(__arg: ..., /) -> ...:
    """Return the singular positional argument unchanged."""
    return __arg


def delete_layout_widgets(layout: QLayout) -> None:
    """Delete all widgets in a layout."""
    while (item := layout.takeAt(0)) is not None:
        item.widget().deleteLater()


def icon_from_bytes(data: bytes) -> QIcon:
    """Create a :py:class:`QIcon` from bytes using a :py:class:`QPixmap` as a proxy."""
    pixmap = QPixmap()
    pixmap.loadFromData(data)
    icon = QIcon(pixmap)
    return icon


# noinspection PyUnresolvedReferences
def init_objects(object_data: dict[QObject, dict[str, Any]], translator: Translator | None = None) -> None:
    """Initialize :py:class:`QObject` attributes with the given data.

    Translation key strings are evaluated using the given translator, if provided.

    widget_data should be a dictionary structured like this::

        {
            widget1: {'text': 'some.translation.key', 'clicked': on_click},
            widget2: {'text': 'some.translation.key', 'size': {'fixed': (None, 400)}},
            widget3: {'text': string_value, 'pasted': on_paste, 'returnPressed': on_return},
            widget4: {'activated': when_activated, 'size': widget_size, 'items': (
                f'The Number "{i}"' for i in range(1, 11)
            )}
        }

    :param object_data: Dictionary containing data used to initialize basic QObject values.
    :param translator: Translator used for translation key evaluation.
    """
    if not translator:
        # If translator is unavailable, return any string/key unchanged.
        translator = _return_arg

    # Initialize widget attributes
    for obj, data in object_data.items():
        special_keys = ('items', 'size', 'text',)
        items, size, text = (data.get(key) for key in special_keys)

        # Find setter method for all non specially-handled keys
        for key, val in data.items():
            if key in special_keys:
                continue  # Skip special keys

            # Check if key is a signal on widget
            # If so, connect it to the given function
            if hasattr(obj, key):
                attribute = getattr(obj, key)
                if isinstance(attribute, SignalInstance):
                    if isinstance(val, Iterable):
                        for slot in val:
                            attribute.connect(slot)
                    else:
                        attribute.connect(val)
                    continue

            # Else call setter to update value
            # Capitalize first character of key
            setter_name: str = f'set{key[0].upper()}{key[1:]}'
            getattr(obj, setter_name)(val)

        # Translate dropdown items
        if items is not None:
            obj.addItems(translator(key) for key in items)

        # Set size
        if size is not None:
            for s_type in ('minimum', 'maximum', 'fixed'):
                if size.get(s_type) is not None:
                    if isinstance(size.get(s_type), QSize):
                        # For PySide6.QtCore.QSize objects
                        getattr(obj, f'set{s_type.title()}Size')(size[s_type])
                    elif isinstance(size.get(s_type), Sequence):
                        # For lists, tuples, etc. Set width and height separately.
                        # None can be used rather than int to skip a dimension.
                        if size.get(s_type)[0]:
                            getattr(obj, f'set{s_type.title()}Width')(size[s_type][0])
                        if size.get(s_type)[1]:
                            getattr(obj, f'set{s_type.title()}Height')(size[s_type][1])

        # Translate widget texts
        if text is not None:
            obj.setText(text)


def scroll_to_top(widget: QTextEdit) -> None:
    """Move text cursor to top of text editor."""
    cursor = widget.textCursor()
    cursor.setPosition(0)
    widget.setTextCursor(cursor)


def set_or_swap_icon(mapping: dict[str, QIcon], key: str, value: QIcon):
    """Given a mapping, replace a QIcon value mapped to the given key with data from another, while keeping the same object references."""
    if key in mapping:
        mapping[key].swap(value)
    else:
        mapping[key] = value



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Custom :py:class:`Exception`'s and excepthook implementation."""

__all__ = (
    'ExceptHookCallable',
    'ExceptionEvent',
    'ExceptionHook',
)

import sys
from collections.abc import Callable
from types import TracebackType
from typing import TypeAlias

from .events import *

ExceptHookCallable: TypeAlias = Callable[[type[BaseException], BaseException, TracebackType], None]


class ExceptionEvent(Event):
    """Event fired when an exception is caught by an :py:class:`ExceptionHook`."""
    __slots__ = ('exception', 'traceback')

    def __init__(self, exception: BaseException, traceback: TracebackType) -> None:
        self.exception: BaseException = exception
        self.traceback: TracebackType = traceback


class ExceptionHook:
    """Object that intercepts :py:class:`Exception`'s and handles them"""

    def __init__(self):
        """Initialize the :py:class:`ExceptionHook` for use in a context manager."""
        self.__old_hook: ExceptHookCallable | None = None
        self.event_bus:  EventBus | None = None

    def __call__(self, type_: type[BaseException], exception: BaseException, traceback: TracebackType) -> None:
        """Called when an exception is raised."""
        # Don't handle BaseExceptions
        if not issubclass(type_, Exception):
            return self.__old_hook(type_, exception, traceback)

        EventBus['exceptions'] << ExceptionEvent(exception, traceback)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} (old_hook={self.__old_hook})>'

    def __enter__(self) -> None:
        """Temporary extend current exception hook."""
        self.__old_hook = sys.excepthook
        self.event_bus = EventBus('exceptions')
        sys.excepthook = self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Reset current exception hook to the original one."""
        sys.excepthook = self.__old_hook
        del EventBus['exceptions']

    @property
    def old_hook(self):
        """The original exception hook."""
        return self.__old_hook



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Main networking package for HaloInfiniteGetter."""

__all__ = (
    'Client',
    'decode_url',
    'dict_to_cookie_list',
    'dict_to_query',
    'encode_url_params',
    'http_code_map',
    'NetworkSession',
)

from .client import Client
from .manager import NetworkSession
from .utils import decode_url
from .utils import dict_to_cookie_list
from .utils import dict_to_query
from .utils import encode_url_params
from .utils import guess_json_utf
from .utils import http_code_map
from .utils import is_error_status
from .utils import query_to_dict



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Package for managing outgoing requests for hi_getter."""

__all__ = (
    'NetworkSession',
)

import datetime
from collections.abc import Callable
from collections.abc import Mapping
from collections.abc import Sequence
from json import dumps as json_dumps
from pathlib import Path
from typing import Any
from typing import TypeAlias

from PySide6.QtCore import *
from PySide6.QtNetwork import *

from ..models import DeferredCallable
from .structures import CaseInsensitiveDict
from .utils import dict_to_query
from .utils import encode_url_params
from .utils import query_to_dict

_NetworkReplyConsumer: TypeAlias = Callable[[QNetworkReply], None]


class NetworkSession:
    """Wrapper for the QNetworkAccessManager."""
    KNOWN_HEADERS: CaseInsensitiveDict[tuple[QNetworkRequest.KnownHeaders, type]] = CaseInsensitiveDict({
        'Content-Disposition':  (QNetworkRequest.ContentDispositionHeader, str),
        'Content-Type':         (QNetworkRequest.ContentTypeHeader, str),
        'Content-Length':       (QNetworkRequest.ContentLengthHeader, bytes),
        'Cookie':               (QNetworkRequest.CookieHeader, QNetworkCookie),
        'ETag':                 (QNetworkRequest.ETagHeader, str),
        'If-Match':             (QNetworkRequest.IfMatchHeader, QStringListModel),
        'If-Modified-Since':    (QNetworkRequest.IfModifiedSinceHeader, QDateTime),
        'If-None-Match':        (QNetworkRequest.IfNoneMatchHeader, QStringListModel),
        'Last-Modified':        (QNetworkRequest.LastModifiedHeader, QDateTime),
        'Location':             (QNetworkRequest.LocationHeader, QUrl),
        'Server':               (QNetworkRequest.ServerHeader, str),
        'Set-Cookie':           (QNetworkRequest.SetCookieHeader, QNetworkCookie),
        'User-Agent':           (QNetworkRequest.UserAgentHeader, str),
    })

    def __init__(self):
        """Initialize the NetworkWrapper."""
        self.manager = QNetworkAccessManager(None)
        self._headers: CaseInsensitiveDict[Any] = CaseInsensitiveDict()

    @property
    def cookies(self) -> dict[str, str]:
        """Dictionary representation of the internal :py:class:`QNetworkCookieJar`."""
        return {cookie.name().toStdString(): cookie.value().toStdString() for cookie in self.manager.cookieJar().allCookies()}

    @cookies.deleter
    def cookies(self) -> None:
        """Clear cookies on delete."""
        self.clear_cookies()

    @property
    def headers(self) -> CaseInsensitiveDict[Any]:
        """Dictionary containing the default session headers."""
        return self._headers

    @headers.setter
    def headers(self, value: Mapping) -> None:
        """Translate any mapping value to a CaseInsensitiveDict for use as headers."""
        if not isinstance(value, Mapping):
            raise TypeError(f'NetworkSession headers must be a Mapping, not {type(value)}')

        self._headers = CaseInsensitiveDict(value)

    @headers.deleter
    def headers(self) -> None:
        """Clear headers on delete."""
        self._headers.clear()

    def _translate_header_value(self, header: str, value: Any):
        old_value = value

        # Match the known-header's value name and translate value to that type.
        match self.KNOWN_HEADERS[header][1].__name__:
            case 'str':
                value = str(old_value)

            case 'bytes':
                if isinstance(old_value, str):
                    value = old_value.encode('utf8')

            case 'QDateTime':
                value = QDateTime()
                if isinstance(old_value, (datetime.datetime, datetime.date, datetime.time)):
                    # Translate datetime objects to a string
                    if not isinstance(old_value, datetime.datetime):
                        date: datetime.date = datetime.datetime.now().date() if isinstance(old_value, datetime.time) else old_value
                        time: datetime.time = datetime.datetime.now().time() if isinstance(old_value, datetime.date) else old_value
                        old_value = datetime.datetime.fromisoformat(f'{date.isoformat()}T{time.isoformat()}')

                    old_value = old_value.isoformat()
                # Translate string to QDateTime object
                value.fromString(str(old_value), Qt.DateFormat.ISODateWithMs)

            case 'QNetworkCookie':
                cookie_list: list[QNetworkCookie] = list()
                match old_value[0] if old_value else None:
                    # Translate dictionaries
                    case Mapping():
                        for cookie in old_value:
                            for name, _value in cookie.items():
                                cookie_list.append(QNetworkCookie(name.encode('utf8'), _value.encode('utf8')))
                    # Translate tuples, lists, etc. that contain two strings (name and value)
                    case Sequence():
                        for cookie in old_value:
                            cookie_list.append(QNetworkCookie(cookie[0].encode('utf8'), cookie[1].encode('utf8')))
                value = cookie_list

            case 'QStringListModel':
                value = [str(item) for item in old_value]

            case 'QUrl':
                if not isinstance(old_value, QUrl):
                    value = QUrl(str(old_value))
        return value

    def clear_cookies(self, domain: str | None = None, path: str | None = None, name: str | None = None) -> bool:
        """Clear some cookies. Functionally equivalent to http.cookiejar.clear.

        Invoking this method without arguments will clear all cookies.  If
        given a single argument, only cookies belonging to that domain will be
        removed.  If given two arguments, cookies belonging to the specified
        path within that domain are removed.  If given three arguments, then
        the cookie with the specified name, path and domain is removed.

        :return: True if a cookie was deleted.
        """

        def deletion_predicate(cookie: QNetworkCookie):
            if name is not None:      # 3 args
                return cookie.name().toStdString() == name and cookie.domain() == domain and cookie.path() == path
            elif path is not None:    # 2 args
                return cookie.domain() == domain and cookie.path() == path
            elif domain is not None:  # 1 arg
                return cookie.domain() == domain
            else:                     # 0 args
                return True

        results = []
        for _cookie in self.manager.cookieJar().allCookies():
            if deletion_predicate(_cookie):
                results.append(self.manager.cookieJar().deleteCookie(_cookie))

        return any(results)

    def set_cookie(self, name: str, value: str, domain: str, path: str | None = None) -> bool:
        """Create a new cookie with the given date.

        Replaces a pre-existing cookie with the same identifier if it exists.

        :return: True if the cookie was set.
        """
        cookie = QNetworkCookie(name=name.encode('utf8'), value=value.encode('utf8'))
        cookie.setDomain(domain)
        cookie.setPath(path or '/')
        return self.manager.cookieJar().insertCookie(cookie)

    def request(self, method: str, url: QUrl | str,
                params: dict[str, str] | list[tuple[str, str]] | None = None,
                data: bytes | dict[str, str] | list[tuple[str, str]] | None = None,
                headers: dict[str, Any] | list[tuple[str, Any]] | None = None,
                cookies: dict[str, str] | list[tuple[str, str]] | None = None,
                # TODO: Finish requests-like implementation
                # files: dict[str, Any] | None = None,
                # auth: tuple[str, str] | None = None,
                # timeout: float | tuple[float, float] | None = None,
                allow_redirects: bool = True,
                # proxies: dict[str, str] | None = None,
                # hooks: dict[str, Callable | Iterable[Callable]] | None = None,
                # stream: bool | None = None,
                verify: bool | str | None = None,
                cert: str | tuple[str, str] | None = None,
                json: dict[str, Any] | None = None,
                finished: _NetworkReplyConsumer | None = None):
        """Send an HTTP request to the given URL with the given data."""

        # Setup values for the request

        url = QUrl(url)
        params = {} if params is None else params
        headers = {} if headers is None else headers
        cookies = {} if cookies is None else cookies

        # Translate tuple pair lists to dictionaries
        # Ex: [('name', 'value'), ('key': 'value')] -> {'name': 'value', 'key': 'value'}
        for tuple_list in ('params', 'data', 'headers', 'cookies'):
            if isinstance(vars()[tuple_list], list):
                vars()[tuple_list] = {key: value for key, value in vars()[tuple_list]}

        request_params = query_to_dict(url.query()) | params  # Override QUrl params with params argument
        request_headers = self.headers.copy() | headers       # Override session headers with headers argument
        request_cookies = self.cookies | cookies              # Override session cookies with cookies argument

        url.setQuery(dict_to_query(request_params))

        # HTTP Body

        content_type = None
        body = None

        if data:
            if isinstance(data, dict):
                body = encode_url_params(data).encode('utf8')
                content_type = 'application/x-www-form-urlencoded'

            elif isinstance(data, bytes):
                body = data

        elif json is not None:
            body = json_dumps(json, allow_nan=False).encode('utf8')
            content_type = 'application/json'

        if content_type and 'Content-Type' not in headers:
            headers['Content-Type'] = content_type

        # Cookies

        original_cookie_jar: QNetworkCookieJar = self.manager.cookieJar()
        if cookies:
            self.manager.setCookieJar(QNetworkCookieJar(self.manager))

            for name, value in request_cookies:
                self.set_cookie(name, value, url.host())

        request = QNetworkRequest(url)

        # SSL Configuration

        ssl_config = QSslConfiguration.defaultConfiguration()

        if isinstance(verify, str):
            ssl_config.setCaCertificates(QSslCertificate.fromPath(verify))

        if isinstance(cert, str):
            ssl_config.setLocalCertificate(QSslCertificate.fromPath(cert))
        elif isinstance(cert, tuple):
            # cert is a tuple of (cert_path, key_path)
            ssl_config.setLocalCertificate(QSslCertificate.fromPath(cert[0]))
            ssl_config.setPrivateKey(QSslKey(Path(cert[1]).read_bytes(), QSsl.Rsa, QSsl.Pem, QSsl.PrivateKey))

        request.setSslConfiguration(ssl_config)

        # Headers

        for name, value in request_headers.items():
            if name in self.KNOWN_HEADERS:
                value = self._translate_header_value(name, value)
                request.setHeader(self.KNOWN_HEADERS[name][0], value)
                continue

            try:
                encoded_value = bytes(value) if not isinstance(value, str) else value.encode('utf8')
            except TypeError:
                encoded_value = str(value).encode('utf8')

            request.setRawHeader(name.encode('utf8'), encoded_value)

        # Other

        if not allow_redirects:
            self.manager.setRedirectPolicy(QNetworkRequest.ManualRedirectPolicy)

        # Handle Reply

        reply: QNetworkReply = self.manager.sendCustomRequest(request, method.encode('utf8'), data=body)

        if verify is False:
            reply.ignoreSslErrors()

        if finished is not None:
            reply.finished.connect(DeferredCallable(finished, reply))

        if self.manager.cookieJar() is not original_cookie_jar:
            self.manager.setCookieJar(original_cookie_jar)
        if self.manager.redirectPolicy() != QNetworkRequest.NoLessSafeRedirectPolicy:
            self.manager.setRedirectPolicy(QNetworkRequest.NoLessSafeRedirectPolicy)

        return reply

    def get(self, url: QUrl | str, **kwargs):
        """Create and send a request with the GET HTTP method."""
        return self.request(method='GET', url=url, **kwargs)

    def head(self, url: QUrl | str, **kwargs):
        """Create and send a request with the HEAD HTTP method."""
        return self.request(method='HEAD', url=url, **kwargs)

    def post(self, url: QUrl | str, **kwargs):
        """Create and send a request with the POST HTTP method."""
        return self.request(method='POST', url=url, **kwargs)

    def put(self, url: QUrl | str, **kwargs):
        """Create and send a request with the PUT HTTP method."""
        return self.request(method='PUT', url=url, **kwargs)

    def delete(self, url: QUrl | str, **kwargs):
        """Create and send a request with the DELETE HTTP method."""
        return self.request(method='DELETE', url=url, **kwargs)

    def patch(self, url: QUrl | str, **kwargs):
        """Create and send a request with the PATCH HTTP method."""
        return self.request(method='PATCH', url=url, **kwargs)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module for the main application classes."""

__all__ = (
    'app',
    'GetterApp',
    'Theme',
)

import json
import sys
from collections.abc import Callable
from collections.abc import Sequence
from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..constants import *
from ..events import *
from ..lang import Translator
from ..models import DeferredCallable
from ..models import DistributedCallable
from ..network import *
from ..tomlfile import *
from ..utils import has_package
from .utils import icon_from_bytes
from .utils import set_or_swap_icon


def app() -> 'GetterApp':
    """:return: GetterApp.instance()"""
    return GetterApp.instance()


class Theme:
    """Object containing data about a Theme."""
    __slots__ = ('id', 'style', 'display_name')

    def __init__(self, id: str, style: str = '', display_name: str | None = None) -> None:
        """Create a new Theme using a unique ID, the style data (.qss file content), and an optional display name.

        :param id: Uniquely identifying string, must be a valid TOML Table name.
        :param style: String read from a .qss file to use as a stylesheet.
        :param display_name: Optional display name to represent the theme, as opposed to the id.
        """
        self.id:           str = id
        self.style:        str = style
        self.display_name: str = display_name if display_name is not None else self.id

        app().themes[self.id] = self


class GetterApp(QApplication):
    """The main HaloInfiniteGetter PySide application that runs in the background and manages the process.

    :py:class:`GetterApp` is a singleton and can be accessed via the class using the GetterApp.instance() method or the app() function.
    """

    # PyCharm detects dict literals in __init__ as a dict[str, EventBus], for no explicable reason.
    # noinspection PyTypeChecker
    def __init__(self, argv: Sequence[str], settings: TomlFile, first_launch: bool = False) -> None:
        """Create a new app with the given arguments and settings."""
        super().__init__(argv)
        self._first_launch: bool = first_launch
        self._legacy_style: str = self.styleSheet()  # Set legacy style before it is overridden
        self._registered_translations: DistributedCallable[set[Callable[DeferredCallable[str]]]] = DistributedCallable(set())

        self.icon_store:      dict[str, QIcon] = {}
        self.session:         NetworkSession = NetworkSession()
        self.settings:        TomlFile = settings
        self.themes:          dict[str, Theme] = {}
        self.theme_index_map: dict[str, int] = {}
        self.translator: Translator = Translator(settings['language'])

        # Register callables to events
        EventBus['settings'] = self.settings.event_bus
        EventBus['settings'].subscribe(DeferredCallable(self.load_themes), TomlEvents.Import)
        EventBus['settings'].subscribe(DeferredCallable(self.update_language), TomlEvents.Import)
        EventBus['settings'].subscribe(DeferredCallable(self.update_stylesheet), TomlEvents.Set, event_predicate=lambda e: e.key == 'gui/themes/selected')

        # Load resources from disk
        self.load_icons()
        self.load_themes()

        # Set the default icon for all windows.
        self.setWindowIcon(self.icon_store['hi'])

    @classmethod
    def instance(cls) -> 'GetterApp':
        """Return the singleton instance of :py:class:`GetterApp`."""
        o: GetterApp | None = super().instance()
        if o is None:
            raise RuntimeError(f'Called {cls.__name__}.instance() when {cls.__name__} is not instantiated.')
        return o

    @property
    def first_launch(self) -> bool:
        """Return whether this is the first launch of the application.

        This is determined by checking if the .LAUNCHED file exists in the user's config folder.
        """
        return self._first_launch

    def _translate_http_code_map(self) -> None:
        """Translate the HTTP code map to the current language."""
        for code in (400, 401, 403, 404, 405, 406):
            http_code_map[code] = (http_code_map[code][0], self.translator(f'network.http.codes.{code}.description'))

    def init_translations(self, translation_calls: dict[Callable, str]) -> None:
        """Initialize the translation of all objects.

        Register functions to call with their respective translation keys.
        This is used to translate everything in the GUI.
        """

        for func, key in translation_calls.items():
            # Call the function with the deferred translation of the given key.
            translate = DeferredCallable(func, DeferredCallable(self.translator, key))

            # Register the object for dynamic translation
            self._registered_translations.callables.add(translate)
            translate()

    def update_language(self) -> None:
        """Set the application language to the one currently selected in settings.

        This method dynamically translates all registered text in the GUI to the given language using translation keys.
        """
        self.translator.language = self.settings['language']
        self._translate_http_code_map()
        self._registered_translations()

    def update_stylesheet(self) -> None:
        """Set the application stylesheet to the one currently selected in settings."""
        self.setStyleSheet(self.themes[self.settings['gui/themes/selected']].style)

    def load_env(self, verbose: bool = True) -> None:
        """Load environment variables from .env file."""
        if not has_package('python-dotenv'):
            dummy_widget = QWidget()
            QMessageBox.critical(
                dummy_widget, self.translator('errors.missing_package.title'),
                self.translator('errors.missing_package.description', 'python-dotenv', Path(sys.executable))
            )
            dummy_widget.deleteLater()
        else:
            from dotenv import load_dotenv
            load_dotenv(verbose=verbose)

    def load_icons(self) -> None:
        """Load all icons needed for the application.

        Fetch locally stored icons from the HI_RESOURCE_PATH/icons directory

        Asynchronously fetch externally stored icons from urls defined in HI_RESOURCE_PATH/external_icons.json
        """
        # Load locally stored icons
        self.icon_store.update({
            filename.with_suffix('').name: QIcon(str(filename)) for
            filename in (HI_RESOURCE_PATH / 'icons').iterdir() if filename.is_file()
        })

        # Load externally stored icons
        external_icon_links: dict[str, str] = json.loads((HI_RESOURCE_PATH / 'external_icons.json').read_text(encoding='utf8'))

        # pylint: disable=cell-var-from-loop
        for key, url in external_icon_links.items():
            # Create a new handler for every key being requested.
            def handle_reply(reply):
                icon = icon_from_bytes(reply.readAll())
                set_or_swap_icon(self.icon_store, key, icon)
                reply.deleteLater()

            app().session.get(url, finished=handle_reply)

    def load_themes(self) -> None:
        """Load all theme locations from settings and store them in self.themes.

        Also set current theme from settings.
        """
        Theme('legacy', self._legacy_style, 'Legacy (Default Qt)')

        for id_, theme in self.settings['gui/themes'].items():
            if not isinstance(theme, dict):
                continue

            theme: dict = theme.copy()
            path: CommentValue | Path = theme.pop('path')
            if isinstance(path, CommentValue):
                path = Path(path.val)
            if path.is_dir():
                search_path = f'hi_theme+{id_}'

                QDir.addSearchPath(search_path, str(path))
                file = QFile(f'{search_path}:stylesheet.qss')
                file.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text)
                theme['id'] = id_
                theme['style'] = QTextStream(file).readAll()

                Theme(**theme)
                file.close()

        # noinspection PyUnresolvedReferences
        self.theme_index_map = {theme_id: i for i, theme_id in enumerate(theme.id for theme in self.sorted_themes())}
        self.update_stylesheet()

    def sorted_themes(self) -> list[Theme]:
        """List of themes sorted by their display name."""
        # noinspection PyTypeChecker
        return sorted(self.themes.values(), key=lambda theme: theme.display_name)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Utils for the HaloInfiniteGetter networking."""

__all__ = (
    'decode_url',
    'dict_to_cookie_list',
    'dict_to_query',
    'encode_url_params',
    'guess_json_utf',
    'http_code_map',
    'is_error_status',
    'query_to_dict',
)

import codecs
from http import HTTPStatus
from urllib.parse import unquote as decode_url
from urllib.parse import urlencode as encode_url_params

from PySide6.QtCore import *
from PySide6.QtNetwork import *

# pylint: disable=not-an-iterable
http_code_map = {status.value: (status.phrase, status.description) for status in HTTPStatus}
for _code, _description in {
        400: 'Your search path has malformed syntax or bad characters.',
        401: 'No permission -- Your API token is most likely invalid.',
        403: 'Request forbidden -- You cannot get this resource with or without an API token.',
        404: 'No resource found at the given location.',
        405: 'Invalid method -- GET requests are not accepted for this resource.',
        406: 'Client does not support the given resource format.',
}.items():
    http_code_map[_code] = (http_code_map[_code][0], _description)
    del _code, _description


def dict_to_cookie_list(cookie_values: dict[str, str]) -> list[QNetworkCookie]:
    """Transforms a name and value pair into a list of :py:class:`QNetworkCookie` objects."""
    return [QNetworkCookie(
        name=name.encode('utf8'),
        value=value.encode('utf8')
    ) for name, value in cookie_values.items()]


def dict_to_query(params: dict[str, str]) -> QUrlQuery:
    """Transforms a param name and value pair into a :py:class:`QUrlQuery` object."""
    query = QUrlQuery()
    query.setQueryItems([(key, value) for key, value in params.items()])
    return query


def query_to_dict(query: QUrlQuery | str) -> dict[str, str]:
    """Translate a query string with the format of QUrl.query() to a dictionary representation."""
    if isinstance(query, QUrlQuery):
        query = query.query()
    query = query.lstrip('?')

    return {} if not query else {
        key: value for key, value in (
            pair.split('=') for pair in query.split('&')
        )
    }


def is_error_status(status: int) -> bool:
    """ Returns True if the HTTP status code is an error status."""
    return 400 <= status < 600


# NOTICE:
#
# Requests
# Copyright 2019 Kenneth Reitz
# Apache 2.0 License


# Null bytes; no need to recreate these on each call to guess_json_utf
_NULL = "\x00".encode("ascii")  # encoding to ASCII for Python 3
_NULL2 = _NULL * 2
_NULL3 = _NULL * 3


def guess_json_utf(data):
    """
    :rtype: str
    """
    # JSON always starts with two ASCII characters, so detection is as
    # easy as counting the nulls and from their location and count
    # determine the encoding. Also detect a BOM, if present.
    sample = data[:4]
    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return "utf-32"  # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return "utf-8-sig"  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return "utf-16"  # BOM included
    null_count = sample.count(_NULL)
    if null_count == 0:
        return "utf-8"
    if null_count == 2:
        if sample[::2] == _NULL2:  # 1st and 3rd are null
            return "utf-16-be"
        if sample[1::2] == _NULL2:  # 2nd and 4th are null
            return "utf-16-le"
        # Did not detect 2 valid UTF-16 ascii-range characters
    if null_count == 3:
        if sample[:3] == _NULL3:
            return "utf-32-be"
        if sample[1:] == _NULL3:
            return "utf-32-le"
        # Did not detect a valid UTF-32 ascii-range character
    return None



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing code related to translation and language files."""

__all__ = (
    'format_value',
    'LANG_PATH',
    'Language',
    'Translator',
)

import json
import re
from contextlib import AbstractContextManager
from contextlib import contextmanager
from pathlib import Path
from string import ascii_letters
from string import digits
from typing import Annotated
from typing import Any
from typing import Final

from .constants import *

LANG_PATH: Final[Path] = HI_RESOURCE_PATH / 'lang'
"""Directory containing language JSON data."""


def format_value(value: str, *args, _language: 'Language' = None) -> str:
    """Format a str with positional arguments.

    You can use {0} notation to refer to a specific positional argument.
    If the notation chars are replaced with the argument, you can no longer use it as a normal positional argument.

    JSON strings ex::

        {
            "a": "{0} is the same as {0} using only one argument.",
            "b": "{0} %s will not work and require 2 arguments",
            "c": "%s {0} is the same as above, where \"%s\" is now the 2nd argument since the 1st is used by \"{0}\""
        }
    """
    list_args:     list = list(args)
    replaced:      set[str] = set()
    pos_param_ref: re.Pattern = re.compile(r'{([1-9]\d*|0)}')
    key_ref:       re.Pattern = re.compile(r'{[\w\d\-.]*}')

    for match in pos_param_ref.finditer(value):
        match = match[0]
        arg_val = args[int(match.strip('{}'))]
        if match not in replaced:
            replaced.add(match)
            value = value.replace(match, arg_val)
            list_args.remove(arg_val)

    value = value % tuple(list_args)

    replaced.clear()
    if _language is not None:
        for match in key_ref.finditer(value):
            match = match[0]
            if match not in replaced:
                replaced.add(match)
                value = value.replace(match, _language.get_raw(match.strip('{}')))

    return value


class Language:
    """Object containing language data. Retrieves stored data with associated RFC 5646 language tags."""

    def __init__(self, primary: Annotated[str, 2] | None = None, region: Annotated[str, 2] | None = None, **kwargs) -> None:
        """Build a new Language object using given sub-tags.

        For more information on RFC 5646, visit https://datatracker.ietf.org/doc/html/rfc5646

        :param primary: The first part of the language code. i.e. [en]-US | Section 2.2.1
        :param region: The (commonly) second part of the language code. i.e. en-[US] | Section 2.2.4

        :keyword ext_lang: Extension to primary subtag | Section 2.2.2
        :keyword script: Script used to write the language | Section 2.2.3
        :keyword variants: A variant of the given language subtag | Section 2.2.5
        :keyword extensions: Extension to the given language, prefixed with a singleton | Section 2.2.6
        :keyword private_use: Any subtag that is not publicly defined | Section 2.2.7
        :type ext_lang: Annotated[str, 3] | None
        :type script: Annotated[str, 4] | None = None
        :type variants: Sequence[str] | None
        :type extensions: Sequence[tuple[Annotated[str, 1], str]] | None
        :type private_use: Sequence[str] | None

        :raises ValueError: If a given subtag is invalid.
        """

        (ext_lang, script, variants, extensions, private_use) = (
            kwargs.pop('ext_lang', None),
            kwargs.pop('script', None),
            kwargs.pop('variants', None),
            kwargs.pop('extensions', None),
            kwargs.pop('private_use', None),
        )

        self._data: dict[str, str] = {}
        self.tag:  str = ''

        sub_tags:  list[str] = []
        err: Exception | None = None

        if primary is None and ext_lang is None and private_use is None:
            err = ValueError('The primary and/or ext_lang and/or private_use subtag must be filled out.')

        if primary is not None:
            primary = primary.lower()

            # RFC 5646 section 2.2.1.1 and 2.2.1.2
            if not primary.isalpha() or not 3 >= len(primary) >= 2:
                err = ValueError(f'Primary language subtag "{primary}" is not valid.')

            sub_tags.append(primary)

        if ext_lang is not None:
            ext_lang = ext_lang.lower()

            # RFC 5646 section 2.2.2.1
            if not ext_lang.isalpha() or not len(ext_lang) == 3:
                err = ValueError(f'Extended language subtag "{ext_lang}" is not valid.')

            sub_tags.append(ext_lang)

        if script is not None:
            script = script.title()

            # RFC 5646 section 2.2.3.2
            if not script.isalpha() or not len(script) == 4:
                err = ValueError(f'Script subtag "{script}" is not valid.')

            sub_tags.append(script)

        if region is not None:
            region = region.upper()

            # ISO 3166-1 or UN M.49
            if not (len(region) == 2 and region.isalpha()) and not (len(region) == 5 and region[:2].isalpha() and region[2:].isnumeric()):
                err = ValueError(f'Region subtag "{region}" is not valid.')

            sub_tags.append(region)

        if variants is not None:
            checked = []
            for variant in variants:

                # RFC 5646 section 2.2.5.4
                if not (variant[0] in ascii_letters and 8 >= len(variant) >= 5) or not (variant[0] in digits and 8 >= len(variant) >= 4):
                    err = ValueError(f'Variant subtag "{variant}" is not valid.')

                # RFC 5646 section 2.2.5.5
                if variant in checked:
                    err = ValueError(f'Variant subtag "{variant}" is repeated.')

                checked.append(variant)
                sub_tags.append(variant)

        if extensions is not None:
            checked = []
            for singleton, extension in extensions:
                extension = extension.lower()

                if not (len(singleton) == 1 and (singleton in digits or singleton in ascii_letters)):
                    err = ValueError(f'Singleton subtag "{singleton}" is not valid.')

                # RFC 5646 section 2.2.6.3
                if singleton in checked:
                    err = ValueError(f'Singleton subtag "{singleton}" is repeated.')

                # RFC 5646 section 2.2.6.5
                if not (extension.isalnum() and 8 >= len(extension) >= 2):
                    err = ValueError(f'Extension subtag "{extension}" is not valid.')

                checked.append(singleton)
                sub_tags.extend((singleton, extension))

        if private_use is not None:
            sub_tags.append('x')
            for i, private_sub_tag in enumerate(private_use):
                private_sub_tag = private_sub_tag.lower()

                # If previous subtag was not a singleton
                if i != 0 and len(private_use[i - 1]) != 1:
                    if len(private_sub_tag) == 4:
                        private_sub_tag = private_sub_tag.title()
                    elif len(private_sub_tag) == 2:
                        private_sub_tag = private_sub_tag.upper()

                if not (private_sub_tag.isalnum() and 8 >= len(private_sub_tag) >= 1):
                    err = ValueError(f'The private subtag "{private_sub_tag}" is not valid.')

                sub_tags.append(private_sub_tag)

        self.tag = '-'.join(sub_tags)

        if err is not None:
            raise ValueError(str(err)[:-1] + f' in language tag "{self.tag}".') from err

        for lang_file in LANG_PATH.iterdir():
            if lang_file.suffix == '.json' and lang_file.with_suffix('').name.lower() == self.tag.replace('-', '_').lower():
                # TODO: Find closest related language file. Ex: en-EN would find en_us.json if en_en.json does not exist.
                new_data = json.loads(lang_file.read_text(encoding='utf8'))
                self._data.update(new_data)
                break

    def __repr__(self) -> str:
        return f'<Language data for {self.tag}>'

    def __getitem__(self, key) -> str:
        return self.get_raw(key)

    @classmethod
    def from_tag(cls, tag: str) -> 'Language':
        """Build a :py:class:`Language` object using a plain string tag.

        Breaks tag into sub-tags and verifies compliance with RFC 5646.
        """
        # TODO: Add functionality

    def get(self, key: str, *args: Any, default: str | None = None) -> str:
        """Get a translation key and format with the given arguments if required.

        :param key: Key to get in JSON data.
        :param args: Formatting arguments; packed into tuple and used with str modulus.
        :param default: Default string if key doesn't exist. (Is not formatted)
        """
        # Default value is key if not overridden
        default = default if default is not None else key
        result: str = self._data.get(key, default)

        if result is not default:
            return format_value(result, *args, _language=self)
        else:
            # Dont format default value
            quote = '"'
            for arg in (f'%{str(arg) if not isinstance(arg, str) else quote + arg + quote}%' for arg in args):
                result += arg

        return result

    def get_raw(self, key: str) -> str:
        """Get raw result, no formatting or defaults.

        :raises KeyError: If key is not in JSON data.
        """
        return self._data[key]


class Translator:
    """Simple class meant to abstract key translation into a function-like state.

    Usage::
        translate = Translator('en-US')
        translate('a.translation.key') -> american value
        translate.language = 'de-GER'
        translate('a.translation.key') -> german value
    """

    def __init__(self, language: Language | str) -> None:
        self._language = self._lang_to_lang(language)

    def __bool__(self) -> bool:
        """Return whether the Translator is available."""
        return True

    def __call__(self, key: str, *args: Any, **kwargs) -> str:
        """Syntax sugar for get_translation."""
        return self.get_translation(key, *args, **kwargs)

    @property
    def language(self) -> Language:
        """Current :py:class:`Language` for this translator."""
        return self._language

    @language.setter
    def language(self, value: Language | str) -> None:
        self._language = self._lang_to_lang(value)

    def get_translation(self, key: str, *args: Any, default: str | None = None) -> str:
        """Get a translation key's value for the current language."""
        return self.language.get(key, *args, default=default)

    @contextmanager
    def as_language(self, language: Language | str) -> AbstractContextManager[None]:
        """Temporarily translate for a specific language using a context manager."""
        # __enter__
        old_lang = self._language
        self.language = language
        yield

        # __exit__
        self._language = old_lang

    @staticmethod
    def _lang_to_lang(language: Language | str) -> Language:
        """Assert that a given value is a :py:class:`Language`."""
        if not isinstance(language, Language):
            language = str(language)  # Stringify non-language object

        if isinstance(language, str):
            # TODO: Move this to Language.from_tag
            language = language.replace(' ', '-').replace('_', '-').strip()
            language = Language(*language.split('-'))  # For basic primary and region subtag compilation (ex: 'en-US' -> primary: 'en', region: 'US')
        return language



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Miscellaneous objects for hi_getter."""

__all__ = (
    'DeferredCallable',
    'DistributedCallable',
)

from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Generator
from typing import Any
from typing import Generic
from typing import TypeAlias
from typing import TypeVar

_CT = TypeVar('_CT', bound=Collection[Callable])  # Bound to Collection of Callables
_PT = TypeVar('_PT')  # Positional Arguments
_KT = TypeVar('_KT')  # Keyword Arguments
_PTCallable: TypeAlias = Callable[[], _PT]  # Which returns _PT
_KTCallable: TypeAlias = Callable[[], _KT]  # Which returns _KT


class _AbstractCallable(ABC):
    """Abstract callable object.

    Child objects are callable as a shortcut to their `run` method.
    """

    def __call__(self, *args, **kwargs) -> Any:
        """Syntax sugar for self.run()"""
        return self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args, **kwargs) -> Any:
        """Execute functionality with the provided arguments and return the result.

        :param args: positional arguments to call with.
        :param kwargs: keyword arguments to call with.
        """
        raise NotImplementedError


class DeferredCallable(_AbstractCallable, Generic[_PT, _KT]):
    """A :py:class:`Callable` with args and kwargs stored for further execution.

    Supports deferred argument evaluation when using :py:class:`Callable`'s as arguments.
    This allows the value of the stored arguments to dynamically change depending on
    when the :py:class:`DeferredCallable` is executed.
    """
    __slots__ = ('_extra_pos_args', 'call_funcs', 'call_types', 'callable', 'args', 'kwargs')

    def __init__(self, __callable: Callable = lambda: None, /, *args: _PT | _PTCallable,
                 _extra_pos_args: int = 0,
                 _call_funcs: bool = True,
                 _call_types: bool = False,
                 **kwargs: _KT | _KTCallable) -> None:
        """Creates a new :py:class:`DeferredCallable`.

        When called, any additional arguments must be expected with _extra_pos_args.
        Any arguments that exceed the extra positional argument limit will be trimmed off.

        :param __callable: Callable to store for later execution.
        :param args: positional arguments to store
        :param _extra_pos_args: Extra positional arguments to expect with self.run.
        :param _call_funcs: Whether to call non-type callables
        :param _call_types: Whether to call class constructors
        :param kwargs: keyword arguments to store
        """
        self._extra_pos_args:   int = _extra_pos_args
        self.call_funcs:        bool = _call_funcs
        self.call_types:        bool = _call_types
        self.callable:          Callable = __callable
        self.args:              tuple[_PT | _PTCallable, ...] = args
        self.kwargs:            dict[str, _KT | _KTCallable] = kwargs

    def __repr__(self) -> str:
        """Represents the :py:class:`DeferredCallable` with the stored callable, args, and kwargs."""
        args, kwargs = self.args, self.kwargs
        return f'<{type(self).__name__} {self.callable} with {args=}, {kwargs=}>'

    def _evaluate_value(self, val: Any) -> Any:
        """Evaluates any callables to their called values.

        :param val: Value to evaluate.
        :return: The called value, if callable.
        """
        return val() if callable(val) and (
                (isinstance(val, type) and self.call_types) or
                (not isinstance(val, type) and self.call_funcs)
        ) else val

    def run(self, *args: Any, **kwargs: Any) -> Any:
        """Run the stored :py:class:`Callable`.

        Takes any additional arguments and temporarily adds to the stored arguments before execution.

        :param args: positional arguments to pass to callable.
        :param kwargs: keyword arguments to pass callable.
        :raises RuntimeError: Internal callable was not expecting the amount of positional arguments given.
        """
        # Add additional arguments from local args
        args = self.args + args[:self._extra_pos_args]  # Trim all arguments that are not expected
        kwargs |= self.kwargs  # PEP 0584

        # Evaluate all callable arguments
        args = tuple(self._evaluate_value(arg) for arg in args)
        kwargs = {key: self._evaluate_value(val) for key, val in kwargs.items()}

        try:
            return self.callable(*args, **kwargs)
        except TypeError as e:
            if ' positional argument but ' in str(e):
                raise RuntimeError(f'{str(e).split(" ", maxsplit=1)[0]} was not expecting additional args, '
                                   f'{type(self).__name__}._extra_call_args may not be set correctly.') from e


class DistributedCallable(_AbstractCallable, Generic[_CT, _PT, _KT]):
    """A :py:class:`Callable` that distributes arguments to all specified callables.

    Supports generic type hinting for the callable collections and arguments. Ex::

        foo: DistributedCallable[set] = DistributedCallable({bar}, 1, 2, 3, four=4)
        foo2: DistributedCallable[list] = DistributedCallable([bar], 1, 2, 3, four=4)

        # Now there's no error when doing

        foo.callables.add(baz)
        # And
        foo2.callables.append(baz)
    """
    __slots__ = ('callables', 'args', 'kwargs')

    def __init__(self, __callables: _CT = (), *args: _PT, **kwargs: _KT) -> None:
        """Creates a new :py:class:`DistributedCallable` with the stored callables, args, and kwargs.

        _CT is a Type Generic containing a :py:class:`Collection` of callables.
        """
        self.callables: _CT = __callables
        self.args:      tuple[_PT, ...] = args
        self.kwargs:    dict[str, _KT] = kwargs

    def __repr__(self) -> str:
        """Represents the :py:class:`DistributedCallable` with the stored callable, args, and kwargs."""
        args, kwargs = self.args, self.kwargs
        return f'<{type(self).__name__} {len(self.callables)} functions with {args=}, {kwargs=}>'

    def run(self, *args, **kwargs) -> tuple[Any]:
        """Run all stored :py:class:`Callable`'s with the given extra arguments.

        :return: The results of each callable, packaged in a tuple.
        """
        return tuple(func(*self.args, **self.kwargs) for func in self.callables)

    def generate(self, *args, **kwargs) -> Generator[Any]:
        """Run all stored :py:class:`Callable`'s with the given extra arguments. Yielding every result.

        :return: A generator yielding the results of each callable.
        """
        for func in self.callables:
            yield func(*args, **kwargs)
